/*------------------------------------------------------------------------------*/
/** 
*  \file  cloud.cpp
*  \brief Definition of class \c CCloud
*  \author Antoine Bouthors 2004-01-09
*/ 
/*------------------------------------------------------------------------------*/


#include "cloud.h"

using namespace Clouds;


/*------------------------------------------------------------------------------
* Constructeur
------------------------------------------------------------------------------*/
CCloud::CCloud( ulong NbBillboards, float BillboardSize )
:	m_BillboardSize( BillboardSize ),
m_ProjAspect( -1 ),
m_ProjFOV( -1 ),
m_ProjZNear( -1 ),
m_ShaftMaxDistance( 2*SimuParams::worldSizeX ),
m_ShaftStepSize( 1 ),
m_ShaftSubdivisionLevel( 40 ),
m_ShaftClipBottom( 2*SimuParams::groundLevel ),
m_ShaftClipTop( /*SimuParams::skyLevel*/0 ),
myDrawQuality(true),
myRunningFlag(true),
myLoopingFlag(false)
{
    m_Shadow.SetTileModeS( CTexture::TileMode_Clamp);
    m_Shadow.SetTileModeT( CTexture::TileMode_Clamp );
    m_Shadow.SetEnvMode( CTexture::EnvMode_Modulate );
    CreateBillboards( NbBillboards );
}


/*------------------------------------------------------------------------------
* Destructeur
------------------------------------------------------------------------------*/
CCloud::~CCloud(){}


/*---------------------------------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------------------------------*/
/*                                         INITIALISATION                                            */
/*---------------------------------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------------------------------*/


/*------------------------------------------------------------------------------
* Fonctions d'initialisation: cration des diffrents billboards
------------------------------------------------------------------------------*/
void CCloud::CreateBillboards( int NbBillboards )
{
    m_Billboards.clear();
    m_Billboards.resize( NbBillboards );
    
    IT_BillbardVector it = m_Billboards.begin();
    ++it;
    
    int i=1;
    while( it != m_Billboards.end() )
    {
        cout << "Creating billboard " << i << endl;
        (*it).Create( float(i) / float(NbBillboards-1), 256 );
        ++i;
        ++it;
    }
    m_NbBillboards = NbBillboards;
}


/*------------------------------------------------------------------------------
* Fonctions d'initialisation: Initialisation du tableau m_Field
* avec des nuages "au hasard"
------------------------------------------------------------------------------*/
void CCloud::Init( ulong XSize, ulong YSize, ulong ZSize )
{
    m_XSize = XSize;
    m_YSize = YSize;
    m_ZSize = ZSize;
    
    initModel();
    initControlState();
    createRandomClouds();
}

/*------------------------------------------------------------------------------
* Fonctions d'initialisation: Cration des couches de shafts
------------------------------------------------------------------------------*/
void CCloud::CreateShafts()
{
    m_Shells.clear();
    printf( "Creating shaft shells...\n" );
    
    for( float d=m_ShaftStepSize ; d<=m_ShaftMaxDistance ; d+=m_ShaftStepSize )
    {
        CShell Shell = CreateShell( d );
        m_Shells.push_back( Shell );
        printf( "%.2f %%\n", d/m_ShaftMaxDistance*100 );
    }
    
}

/*------------------------------------------------------------------------------
* Fonctions d'initialisation: Cration des couches de shafts
------------------------------------------------------------------------------*/
CShell CCloud::CreateShell( float Distance )
{
    CShell Result;
    float Top = sin( m_ProjFOV*3.1415/360. )*1.1;
    float Bottom = -Top;
    float Right = Top*m_ProjAspect;
    float Left = -Right;
    float XStep = (Right-Left)/m_ShaftSubdivisionLevel;
    float YStep = (Top-Bottom)/m_ShaftSubdivisionLevel;
    
    float y = Bottom;
    int i, j;
    
    for( j=0 ; j<m_ShaftSubdivisionLevel ; ++j )
    {
        float x = Left;
        for( i=0 ; i<m_ShaftSubdivisionLevel ; ++i )
        {
            CPoint BottomRight( x+XStep,  y,  -m_ProjZNear );
            CPoint BottomLeft( x,  y,  -m_ProjZNear );
            CPoint TopRight( x+XStep,  y+YStep,  -m_ProjZNear );
            CPoint TopLeft( x, y+YStep,  -m_ProjZNear );
            BottomRight.Normalize();
            BottomLeft.Normalize();
            TopRight.Normalize();
            TopLeft.Normalize();
            Result.Triangles.push_back( CTriangle( BottomRight, BottomLeft, TopLeft ) );
            Result.Triangles.push_back( CTriangle( BottomRight, TopLeft, TopRight  ) );
            x += XStep;
        }
        y += YStep;        
    }
    
    for( IT_TriangleVector it = Result.Triangles.begin() ; 
         it != Result.Triangles.end();
         ++it )
    {
        (*it).a *= Distance;		
        (*it).b *= Distance;
        (*it).c *= Distance;
    }
    Result.Distance = Distance;
    
    return Result;
}



/*------------------------------------------------------------------------------
* Fonctions d'initialisation: Cration des couches de shafts
------------------------------------------------------------------------------*/
void CCloud::SubdivideShell( const CShell& Old, CShell& New )
{
    New.Triangles.clear();
    
    T_TriangleVector& NewTriangles = New.Triangles;
    const T_TriangleVector& OldTriangles = Old.Triangles;
    CIT_TriangleVector it = OldTriangles.begin();
    
    while( it != OldTriangles.end() )
    {
        const CTriangle& Tri = (*it);
        CPoint a = ( Tri.a + Tri.c )/2;
        CPoint b = ( Tri.a + Tri.b )/2;
        CPoint c = ( Tri.b + Tri.c )/2;
        
        a.Normalize();
        b.Normalize();
        c.Normalize();
        
        CTriangle NewTri1( Tri.a, b, a );
        CTriangle NewTri2( b, Tri.b, c );
        CTriangle NewTri3( a, b, c );
        CTriangle NewTri4( a, c, Tri.c );
        
        NewTriangles.push_back( NewTri1 );
        NewTriangles.push_back( NewTri2 );
        NewTriangles.push_back( NewTri3 );
        NewTriangles.push_back( NewTri4 );
        
        ++it;
    }
}


/*---------------------------------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------------------------------*/
/*                                             RENDU                                                 */
/*---------------------------------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------------------------------*/


/*------------------------------------------------------------------------------
* Cration des billboards lisss  partir du m_Field
* Les billboards sont stocks dans la DistMapSun (sans leur couleur)
------------------------------------------------------------------------------*/
void CCloud::SmoothField( int SmoothSize )
{
    char Bill;
    CPoint CDistToSun;
    float distToSun;
    m_DistMapFromSun.clear();
    for( int k=SmoothSize ; k<m_ZSize-SmoothSize ; ++k )
    {
        for( int j=SmoothSize ; j<m_YSize-SmoothSize ; ++j )
        {
            for( int i=SmoothSize ; i<m_XSize-SmoothSize ; ++i )
            {
                float Val = 0;
                for( int z = k-SmoothSize ; z <= k+SmoothSize ; ++z )
                {
                    for( int y = j-SmoothSize ; y <= j+SmoothSize ; ++y )
                    {
                        for( int x = i-SmoothSize ; x <= i+SmoothSize ; ++x )
                        {
                            //Val += m_Field.Access( x, y, z );
                            if(cloudModel[currentCloudModel].Access(x,y,z).cld)
                                Val += .5;
                            if(cloudModel[currentCloudModel].Access(x,y,z).cld)
                                Val += .5;
                            // printf("on en a un autre!");
                        }
                        
                    }
                }
                Val /= ( 2*SmoothSize+1 )*( 2*SmoothSize+1 )*( 2*SmoothSize+1 );
                Bill = min( m_NbBillboards-1, (ulong)( Val*m_NbBillboards ) );
                if(Bill>0){
                    CDistToSun = CPoint( i-SimuParams::myLight->getX(),
                                         j-SimuParams::myLight->getY(),
                                         k-SimuParams::myLight->getZ());
                    distToSun = CDistToSun.Norm();
                    CElement es; es.x = i; es.y = j; es.z = k;
                    es.Billboard = Bill;
                    m_DistMapFromSun.insert( T_DistMap::value_type( distToSun, es ) );
                    
                }
            }
        }
    }
}


/*------------------------------------------------------------------------------
* RENDU depuis le soleil, et remplissage de la distmap depuis l'oeir
------------------------------------------------------------------------------*/
void CCloud::Shade()
{
    float Projection[4][4];
    float ModelView[4][4];
    
    
    glPushAttrib(GL_LIGHTING_BIT|GL_TEXTURE_BIT|GL_COLOR_BUFFER_BIT|GL_ENABLE_BIT|GL_VIEWPORT_BIT);
    glDisable( GL_LIGHTING );
    glDisable( GL_DEPTH_TEST );
    
    //	Efface l'cran en BLANC
    glClearColor( 1, 1, 1, 1 );
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
    
    //	Projection orthonorme
    glMatrixMode( GL_PROJECTION );
    glPushMatrix();
    glLoadIdentity();
    glOrtho( -m_XSize/1.5, m_XSize/1.5, -m_ZSize/1.5, m_ZSize/1.5, 10, 1000 );
    
    //	Placement de la scne en fonction de la position du soleil
    glMatrixMode( GL_MODELVIEW );
    glPushMatrix();
    glLoadIdentity();
    gluLookAt( SimuParams::myLight->getX(),
               SimuParams::myLight->getY(),
               SimuParams::myLight->getZ(),
               m_XSize/2, m_YSize/2, m_ZSize/2, 0 , 1, 0 );
    
    //	Fonction de blending approprie : la couleur du framebuffer est multiplie
    //	par l'alpha du billboard. La couleur du billboard est nulle. Rsultat : le 
    //	billboard est ajout en "ngatif" dans le framebuffer ( plus l'alpha du
    //	billboard est faible et plus le rsultat sera sombre )
    glEnable( GL_BLEND );
    glBlendFunc( GL_ZERO, GL_SRC_ALPHA );
    
    //	Trouve la rgion la plus grande possible tenant dans la fentre
    //	et donc les largeur et hauteur sont des puissances de 2
    //	Comme a notre texture d'ombre sera directement le framebuffer 
    int GoodWidth = 1;
    int GoodHeight = 1;
    while( GoodWidth*2 <= SimuParams::width ) GoodWidth *=2;
    while( GoodHeight*2 <= SimuParams::height ) GoodHeight *=2;
    glViewport( 0, 0, GoodWidth, GoodHeight );
    
    //	Rcupre la matrice du soleil 
    StoreSunMatrix();
    
    //	Rcupre la matrice de projection
    glGetFloatv( GL_PROJECTION_MATRIX, &Projection[0][0] );
    
    //	Dessine les billboards du plus proche au plus loin du soleil
    glColor4f( 1, 1, 1, 1 );
    glPushMatrix();
    T_DistMap::iterator cit = m_DistMapFromSun.begin();
    
    CPoint CDistToCamera;
    float distToCamera;
    m_DistMapFromEye.clear();
    
    
    
    while( cit != m_DistMapFromSun.end() )
    {
        CElement& Elt = (*cit).second;
        
        glPushMatrix();
        glTranslatef( (float)Elt.x, (float)Elt.y, (float)Elt.z );
        
        //		"Drotationne" la matrice modelview
        glGetFloatv( GL_MODELVIEW_MATRIX, &ModelView[0][0] );
        ModelView[0][0] = ModelView[1][1] = ModelView[2][2] = 1;
        ModelView[0][1] = ModelView[0][2] = 0;
        ModelView[1][0] = ModelView[1][2] = 0;
        ModelView[2][0] = ModelView[2][1] = 0;
        glLoadMatrixf( &ModelView[0][0] );
        
        //		Trouve le pixel dans la fentre correspondant au centre du billboard
        //		Les quations matricielles sont simplifies du fait que les coordonnes
        //		du centre du billboard sont (0,0,0)
        float x = ModelView[3][0]*Projection[0][0]
            + ModelView[3][1]*Projection[1][0]
            + ModelView[3][2]*Projection[2][0]
            + Projection[3][0]; 
        float y = ModelView[3][0]*Projection[0][1]
            + ModelView[3][1]*Projection[1][1]
            + ModelView[3][2]*Projection[2][1]
            + Projection[3][1];
        
        int r = (int)GoodWidth/2 * ( 1 + x );
        int s = (int)GoodHeight/2 * ( 1 + y );
        
        //		Lit la valeur du pixel et l'enregistre comme couleur du billboard
        //
        float Pixel;
        glReadPixels( r, s, 1, 1, GL_RED, GL_FLOAT, &Pixel );
        
        CDistToCamera = CPoint( Elt.x-SimuParams::myCamera->getX(),
                                Elt.y-SimuParams::myCamera->getY(),
                                Elt.z-SimuParams::myCamera->getZ());
        distToCamera = CDistToCamera.Norm();
        CElement ec = Elt; ec.Color = .2f + Pixel*.8f;
        m_DistMapFromEye.insert( T_DistMap::value_type( distToCamera, ec ) );
        
        //		Dessine le billboard
        m_Billboards[ Elt.Billboard ].Draw( m_BillboardSize );
        //		Vide les buffers pour que le billboard soit dessin
        glFinish(); 
        
        glPopMatrix();
        ++cit;
    }
    glPopMatrix();
    
    //	Charge la texture d'ombre
    glFinish();
    m_Shadow.LoadFromScreen();
    
    glMatrixMode( GL_PROJECTION );
    glPopMatrix();
    glMatrixMode( GL_MODELVIEW );
    glPopMatrix();
    
    glPopAttrib();
}

/**
*Re-tri de la DistMapEye
 */
void CCloud::sortDistMapEye(){
    CPoint CDistToCamera;
    float distToCamera;
    T_DistMap tempDistMap;
    T_DistMap::iterator cit = m_DistMapFromEye.begin();
    tempDistMap.clear();
    while( cit != m_DistMapFromEye.end() )
    {
        CElement& Elt = (*cit).second;
        CDistToCamera = CPoint( Elt.x-SimuParams::myCamera->getX(),
                                Elt.y-SimuParams::myCamera->getY(),
                                Elt.z-SimuParams::myCamera->getZ());
        distToCamera = CDistToCamera.Norm();
        ++cit;
        tempDistMap.insert( T_DistMap::value_type( distToCamera, Elt ) );
    }
    m_DistMapFromEye.clear();
    m_DistMapFromEye = tempDistMap;
}




/*------------------------------------------------------------------------------
Rendu complet depuis l'oeil
------------------------------------------------------------------------------*/
void CCloud::Draw()
{
    
    glPushAttrib( GL_LIGHTING_BIT
                  | GL_TEXTURE_BIT
                  | GL_COLOR_BUFFER_BIT
                  | GL_ENABLE_BIT
                  | GL_DEPTH_BUFFER_BIT );
    //	Fonction de blending standard : dessine les billboards en les faisant
    //	s'ajouter les uns sur les autres
    glEnable( GL_BLEND );
    
    //	Utilise le depth test mais n'crit rien dans le zbuffer
    glDepthMask( GL_FALSE );
    glDisable( GL_LIGHTING );
    
    
    glBlendFunc( GL_ONE, GL_SRC_ALPHA );
    
    
    //	Dessine les billboards et les shells du plus loign vers le plus proche du 
    //	point de vue
    T_ShellVector::reverse_iterator sit = m_Shells.rbegin();
    
    if(m_DistMapFromEye.size() != 0){
        
        T_DistMap::reverse_iterator dit = m_DistMapFromEye.rbegin();
        
        //Skip the far shells
        if(myDrawQuality){
            while( sit != m_Shells.rend() && (*sit).Distance > (*dit).first ){ //
                RenderShell( *sit );
                ++sit;
            }
        }
        
        while( dit != m_DistMapFromEye.rend() )
        {
            float BillboardDistance = (*dit).first;
            
            if(myDrawQuality){
                while( sit != m_Shells.rend() && (*sit).Distance > BillboardDistance ) { 
                    RenderShell( *sit );
                    ++sit;
                }
            }
            
            CElement& Elt = (*dit).second;
            glColor3f( 1, 1, 1 );
            
            glPushMatrix();
            glTranslatef( (float)Elt.x, (float)Elt.y, (float)Elt.z );
            //		"Drotationne" la matrice modelview
            UnRotateModelViewMatrix();
            //		Applique la couleur du billboard trouve dans Shade()
            glColor4f( Elt.Color, Elt.Color, Elt.Color, 1 );
            //		Dessine le billboard
            m_Billboards[ Elt.Billboard ].Draw( m_BillboardSize );
            glPopMatrix();
            
            ++dit;
        }
    }
    //Skip the near shells
    if(myDrawQuality){
        while( sit != m_Shells.rend() ){
            RenderShell( *sit );
            ++sit;
        }
    }
    
    glPopAttrib();
}



/*------------------------------------------------------------------------------
* Rendu des shafts
------------------------------------------------------------------------------*/
void CCloud::RenderShell( const CShell& Shell )
{
    CMatrix ModelView;
    ModelView.LoadFromModelview();
    ModelView.Inverse();
    
    glMatrixMode( GL_MODELVIEW );
    glPushMatrix();
    glLoadIdentity();
    glBlendFunc( GL_ONE, GL_ONE );
    
    m_Shadow.Bind();
    const float DeltaS = .005;
    const float SunColor[3] = { .9f, .9f, .7f };
    const float SunColorDelta[3] = {SunColor[0]*DeltaS,SunColor[1]*DeltaS,SunColor[2]*DeltaS};
    CPoint TexCoords[3] = {
        CPoint( 0, 0, 0 ),
        CPoint( 0, 0, 0 ),
        CPoint( 0, 0, 0 )
    };
    CPoint Transformed[3];
    
    for( CIT_TriangleVector it = Shell.Triangles.begin(); it != Shell.Triangles.end() ; ++it )
    {	
        const CTriangle& Tri = (*it);
        Transformed[0] = ModelView*Tri.a;
        Transformed[1] = ModelView*Tri.b;
        Transformed[2] = ModelView*Tri.c;
        
        float Height[3] = {
            AtmFunction( Transformed[0].y ),
            AtmFunction( Transformed[1].y ),
            AtmFunction( Transformed[2].y )
        };
        
        float Color[3][3] = {
        { SunColorDelta[0]*Height[0], SunColorDelta[1]*Height[0], SunColorDelta[2]*Height[0] },
        { SunColorDelta[0]*Height[1], SunColorDelta[1]*Height[1], SunColorDelta[2]*Height[1] },
        { SunColorDelta[0]*Height[2], SunColorDelta[1]*Height[2], SunColorDelta[2]*Height[2] }
        };
        
        for(short i=0;i<3;i++)
            TexCoords[i].x = TexCoords[i].y = TexCoords[i].z = 0;
        
        if(    Transformed[0].y < m_ShaftClipBottom 
               || Transformed[1].y < m_ShaftClipBottom 
               || Transformed[2].y < m_ShaftClipBottom )		
            continue;
        
        if(    Transformed[0].y < m_ShaftClipTop 
               && Transformed[1].y < m_ShaftClipTop 
               && Transformed[2].y < m_ShaftClipTop )
        {
            GetShadowTexCoord( Transformed[0], TexCoords[0] );
            GetShadowTexCoord( Transformed[1], TexCoords[1] );
            GetShadowTexCoord( Transformed[2], TexCoords[2] );
        }
        glBegin( GL_TRIANGLES );		 
        glColor3fv( Color[0] ); glTexCoord2fv( (const float*)TexCoords[0] ); glVertex3fv( (const float*)Tri.a );
        glColor3fv( Color[2] ); glTexCoord2fv( (const float*)TexCoords[2] ); glVertex3fv( (const float*)Tri.c );
        glColor3fv( Color[1] ); glTexCoord2fv( (const float*)TexCoords[1] ); glVertex3fv( (const float*)Tri.b );
        glEnd();
    }
    
    glBlendFunc( GL_ONE, GL_SRC_ALPHA );
    glPopMatrix();
}




/*------------------------------------------------------------------------------
* Enregistre la matrice du soleil (modelview * projection) pour pouvoir 
* l'utiliser ensuite pour trouver les coordonnes de texture pour le sol
* (ombre) et les shafts of lights
------------------------------------------------------------------------------*/
void CCloud::StoreSunMatrix()
{
    CMatrix Projection;
    CMatrix ModelView;
    Projection.LoadFromProjection();
    ModelView.LoadFromModelview();
    
    m_SunMatrix = ModelView*Projection;
}


/*------------------------------------------------------------------------------
* Trouve les coordonnes de la texture d'ombre  partir des matrices
* du soleil et du point sur lequel projeter
------------------------------------------------------------------------------*/
void CCloud::GetShadowTexCoord( const CPoint& p, CPoint& t )
{
    t = m_SunMatrix*p;
    
    
    t.x = ( t.x + 1 ) /2.f;
    t.y = ( t.y + 1 ) /2.f;
}

/*---------------------------------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------------------------------*/
/*                                              MODEL                                                */
/*---------------------------------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------------------------------*/
int rnd(){
    return rand()%100;
}

void CCloud::initModel(){
    cloudModel[0].Init(m_XSize,m_YSize,m_ZSize);
    cloudModel[1].Init(m_XSize,m_YSize,m_ZSize);    
    extProbField.Init(m_XSize,m_YSize,m_ZSize);
    humProbField.Init(m_XSize,m_YSize,m_ZSize);
    actProbField.Init(m_XSize,m_YSize,m_ZSize);    
    cloudpoint p(false,false,false);
    cloudModel[0].Fill(p);
    cloudModel[1].Fill(p);
    extProbField.Fill(100);
    humProbField.Fill(0);
    actProbField.Fill(0);
    currentCloudModel = 0;
    frameWindCounter = m_YSize;
    //default KeyboardMode compatibility
    setDefaultModeCompatibility(false);
}

void CCloud::createRandomClouds(){
//    extProbField.Fill(20);
//    humProbField.Fill(0);
//    actProbField.Fill(0);
    
    int numClouds = 8 + random()%3;
    for (int idCloud = 0;idCloud<numClouds;idCloud++){
        int cloudSizeX = 1 + random()%(m_XSize/8);
        int cloudSizeY = 1 + random()%(m_YSize/8);
        int cloudSizeZ = 1 + random()%(m_ZSize/8);
        int cloudCenterX = cloudSizeX + random()%(m_XSize-1-2*cloudSizeX);
        int cloudCenterY = cloudSizeY + random()%(m_YSize-1-2*cloudSizeY);
        int cloudCenterZ = cloudSizeZ + random()%(m_ZSize-1-2*cloudSizeZ);
        for (int x = -cloudSizeX; x <= cloudSizeX; x++){
            for (int y = -cloudSizeY;y <= cloudSizeY; y++){
                for (int z = -cloudSizeZ;z <= cloudSizeZ; z++){
                    float xweight = (float)x/(float)cloudSizeX;
                    float yweight = (float)y/(float)cloudSizeY;
                    float zweight = (float)z/(float)cloudSizeZ;
                    float ratio = 6 * sqrt(xweight*xweight+yweight*yweight+zweight*zweight) + 1;
                 //   printf("x: %i y: %i z: %i xw: %f yw: %f zw: %f ratio: %f act: %f ext: %f hum: %f\n",x,y,z,xweight,yweight,zweight,ratio,100.0f/ratio,ratio*4.0f,80.0f/ratio);
                    actProbField.Access(cloudCenterX+x,cloudCenterY+y,cloudCenterZ+z) = 100.0f/(float)ratio;
                    extProbField.Access(cloudCenterX+x,cloudCenterY+y,cloudCenterZ+z) = ((float)ratio)*4.0f;
                    humProbField.Access(cloudCenterX+x,cloudCenterY+y,cloudCenterZ+z) = 80.0f/(float)ratio;
                }
            }
        }
//        printf("\n\n");
    }

/*    for(int i=0;i<m_XSize;i++)
        for(int j=0;j<m_YSize;j++)
            for(int k=0;k<m_ZSize;k++){
                if( i>=2*(m_XSize/20) && i<=4*(m_XSize/20) &&
                    j>=(m_YSize/5) && j<=4*(m_YSize/5)     && (
                                                     ( k>=3*(m_ZSize/4) && k<=m_ZSize ) ||
                                                     ( k>=0 && k<=(m_ZSize/4) )
                                                     )
                    ){
                    actProbField.Access(i,j,k) = 55;
                    extProbField.Access(i,j,k) = 75;
                    humProbField.Access(i,j,k) = 55;
                } else {
                    actProbField.Access(i,j,k) = 0;
                    extProbField.Access(i,j,k) = 1;
                    humProbField.Access(i,j,k) = 0;
                }
            }
  */              
}

bool CCloud::activeNear(int i,int j, int k, int modelID){
    return cloudModel[modelID].Access(i+1,j,k).act ||
    cloudModel[modelID].Access(i,j+1,k).act ||
    cloudModel[modelID].Access(i,j,k+1).act ||
    cloudModel[modelID].Access(i-1,j,k).act ||
    cloudModel[modelID].Access(i,j-1,k).act ||
    cloudModel[modelID].Access(i,j,k-1).act ||
    cloudModel[modelID].Access(i+2,j,k).act ||
    cloudModel[modelID].Access(i,j+2,k).act ||
    cloudModel[modelID].Access(i-2,j,k).act ||
    cloudModel[modelID].Access(i,j-2,k).act ||
    cloudModel[modelID].Access(i,j,k-2).act;
}

int CCloud::windVelocity(int k){
    return 1+k/m_YSize;
    //return (rand()%m_YSize)<k;
}
void CCloud::update(){
    frameWindCounter= (frameWindCounter+1)%5;
    if(myRunningFlag){
        int v;
        int old_i;
        cloudpoint nullp(false,false,false);
        for(unsigned short j=0 ; j<m_YSize ; ++j )
        {
            for(unsigned short k=0 ; k<m_ZSize ; ++k )
            {
                for(unsigned short i=0 ; i<m_XSize ; ++i )
                {
                    v=windVelocity(j);
                    //printf("counter: %i alt: %i windvel: %i\n",frameWindCounter,j,v);
                    old_i = i-v;
                    if(myLoopingFlag && old_i < 0)
                        old_i += m_XSize;
                    if( old_i > 0 ){
                        if(cloudModel[!currentCloudModel].Access(old_i,j,k).cld)
                            cloudModel[currentCloudModel].Access(i,j,k).cld = (rnd()>=extProbField.Access(i,j,k));
                        else
                            cloudModel[currentCloudModel].Access(i,j,k).cld = cloudModel[!currentCloudModel].Access(old_i,j,k).act;                        
                        if(cloudModel[!currentCloudModel].Access(old_i,j,k).hum)
                            cloudModel[currentCloudModel].Access(i,j,k).hum = !cloudModel[!currentCloudModel].Access(old_i,j,k).act;
                        else
                            cloudModel[currentCloudModel].Access(i,j,k).hum = (rnd()<humProbField.Access(i,j,k));
                        
                        if( cloudModel[!currentCloudModel].Access(old_i,j,k).act )
                            cloudModel[currentCloudModel].Access(i,j,k).act = 
                                cloudModel[!currentCloudModel].Access(i,j,k).act &&
                                cloudModel[!currentCloudModel].Access(old_i,j,k).act &&
                                activeNear(old_i,j,k,!currentCloudModel);
                        else
                            cloudModel[currentCloudModel].Access(i,j,k).act = (rnd()<actProbField.Access(i,j,k));
                    } else {
                        cloudModel[currentCloudModel].Access(i,j,k) = nullp;
                    }
                }
            }
        }
        for( short i=m_XSize-1;i>=0;i--){
            for( short j=m_YSize-1;j>=0;j--){
                for( short k=m_ZSize-1;k>=0;k--){
                    v=windVelocity(j);
                    old_i = i-v;
                    if(myLoopingFlag && old_i < 0)
                        old_i += m_XSize;
                    if( old_i>0 ){
                        actProbField.Access(i,j,k) = actProbField.Access(old_i,j,k);
                        extProbField.Access(i,j,k) = extProbField.Access(old_i,j,k);
                        humProbField.Access(i,j,k) = humProbField.Access(old_i,j,k);
                    } else {
                        extProbField.Access(i,j,k) = 100;
                        actProbField.Access(i,j,k) = 0;
                        humProbField.Access(i,j,k) = 0;
                    }
                }
            }
        }
        SmoothField(1);
        Shade();
        currentCloudModel = !currentCloudModel;
    }//end if myRunningFlag
    else {
        if(SimuParams::myLight->hasMoved()){
            SmoothField(1);
            Shade();
        }
        else if(SimuParams::myCamera->hasMoved())
            sortDistMapEye();
    }

}


/*---------------------------------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------------------------------*/
/*                                            USER CONTROL                                           */
/*---------------------------------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------------------------------*/
#define CLOUD_RENDERSHAFTS_MENUENTRY 0
#define CLOUD_MODELRESET_MENUENTRY 1
#define CLOUD_MODELREGENERATE_MENUENTRY 10
#define CLOUD_MODELPAUSE_MENUENTRY 2
#define CLOUD_MODELLOOP_MENUENTRY 3
#define CLOUD_KBCONTROL_MENUENTRY 4
void CCloud::initControlState(){
    cloudMenu = new Menu("Nuage");
    StaticMenu::currentMenu->addSubMenu(cloudMenu);
//    Menu* modelMenu = new Menu("Modele");
    cloudMenu->addSubMenu(new Menu("Affichage des shafts",this,CLOUD_RENDERSHAFTS_MENUENTRY));
    cloudMenu->addSubMenu(new Menu("Reset",this,CLOUD_MODELRESET_MENUENTRY));
    cloudMenu->addSubMenu(new Menu("Regenerer",this,CLOUD_MODELREGENERATE_MENUENTRY));
    cloudMenu->addSubMenu(new Menu("Pause",this,CLOUD_MODELPAUSE_MENUENTRY));
    cloudMenu->addSubMenu(new Menu("Boucle",this,CLOUD_MODELLOOP_MENUENTRY));
//    cloudMenu->addSubMenu(modelMenu);
    cloudMenu->addSubMenu(new Menu("Controler au clavier",this,CLOUD_KBCONTROL_MENUENTRY));
}

void CCloud::menuItemClicked(int menuEntry){
    switch(menuEntry){
        case CLOUD_RENDERSHAFTS_MENUENTRY :
            if(myDrawQuality){
                this->myDrawQuality = false;
                hud::currentHud->addMessage(new TemporizedMessage("Affichage rapide"));
            }else{
                this->myDrawQuality = true;
                hud::currentHud->addMessage(new TemporizedMessage("Affichage des reflets du soleil active"));                
            }
            break;
        case CLOUD_MODELRESET_MENUENTRY :
            hud::currentHud->addMessage(new TemporizedMessage("Nuages reinitialises"));
            initModel();
            break;
        case CLOUD_MODELREGENERATE_MENUENTRY :
            hud::currentHud->addMessage(new TemporizedMessage("Nouveaux nuages ajoutes"));
            createRandomClouds();
            break;
        case CLOUD_MODELPAUSE_MENUENTRY :
            if(myRunningFlag){
                this->myRunningFlag = false;
                hud::currentHud->addMessage(new TemporizedMessage("Modele en pause"));
            }else{
                this->myRunningFlag = true;
                hud::currentHud->addMessage(new TemporizedMessage("Modele en marche"));
            }
            break;
        case CLOUD_MODELLOOP_MENUENTRY :
            if(myLoopingFlag){
                this->myLoopingFlag = false;
                hud::currentHud->addMessage(new TemporizedMessage("Modele en lineaire"));
            }else{
                this->myLoopingFlag = true;
                hud::currentHud->addMessage(new TemporizedMessage("Modele en boucle"));
            }
            break;
        case CLOUD_KBCONTROL_MENUENTRY :
            hud::currentHud->addMessage(new TemporizedMessage("Controles:"));            
            hud::currentHud->addMessage(new TemporizedMessage("q: affichage des traits de lumiere"));            
            hud::currentHud->addMessage(new TemporizedMessage("r: reset"));            
            hud::currentHud->addMessage(new TemporizedMessage("g: regeneration"));            
            hud::currentHud->addMessage(new TemporizedMessage("p: pause"));            
            hud::currentHud->addMessage(new TemporizedMessage("o: modele en boucle"));            
            Keyboard::selectKeyboardMode(this);
            break;
    }
}

void CCloud::keyboardFunc(unsigned char c, int x, int y){
    switch(c){
        case 'q':
            if(myDrawQuality){
                this->myDrawQuality = false;
                hud::currentHud->addMessage(new TemporizedMessage("Affichage rapide"));
            }else{
                this->myDrawQuality = true;
                hud::currentHud->addMessage(new TemporizedMessage("Affichage des reflets du soleil active"));                
            }
            break;
        case 'r':
            hud::currentHud->addMessage(new TemporizedMessage("Nuages reinitialises"));
            initModel();
            break;
        case 'g':
            hud::currentHud->addMessage(new TemporizedMessage("Nouveaux nuages ajoutes"));
            createRandomClouds();
            break;
        case 'p':
            if(myRunningFlag){
                this->myRunningFlag = false;
                hud::currentHud->addMessage(new TemporizedMessage("Modele en pause"));
            }else{
                this->myRunningFlag = true;
                hud::currentHud->addMessage(new TemporizedMessage("Modele en marche"));
            }
            break;
        case 'o':
            if(myLoopingFlag){
                this->myLoopingFlag = false;
                hud::currentHud->addMessage(new TemporizedMessage("Modele en lineaire"));
            }else{
                this->myLoopingFlag = true;
                hud::currentHud->addMessage(new TemporizedMessage("Modele en boucle"));
            }
            break; 
            case 27:
                Keyboard::selectDefaultKeyboardMode();
        default:
            break;        
    }
}