/*------------------------------------------------------------------------------*/
/** 
 *  \file  Texture.h
 *  \brief Definition of class \c CTexture
 *  \author Antoine Bouthors 2002-10-31
 */ 
/*------------------------------------------------------------------------------*/

#ifndef _Texture_h_
#define _Texture_h_

#include "config.h"

#ifdef _USE_OPENIL_
	#include "il/il.h"
	#include "il/ilu.h"
	#include "il/ilut.h"
#endif // _USE_OPENIL_

namespace Clouds {

/*------------------------------------------------------------------------------*/
/** 
 *  \class  CTexture
 *  \brief  Texture class
 *  \author Antoine Bouthors 2002-10-31
 */ 
/*------------------------------------------------------------------------------*/

class CTexture
{

public:

	enum TTileMode { TileMode_Tile, TileMode_Clamp, TileMode_ClampToBorder };
	enum TFiltering { Filtering_Bilinear, Filtering_Trilinear, Filtering_Anisotropic };
	enum TEnvMode { EnvMode_Replace, EnvMode_Modulate, EnvMode_Decal };

    CTexture();
    CTexture( const uchar* File );
	CTexture( const CTexture& Dup );
	const CTexture& operator=( const CTexture& Dup );
    virtual ~CTexture();

    //-------------------------------------------------------------------------
    /** \name (Un)Loading functions */
    //-------------------------------------------------------------------------
    //@{
	bool LoadFromScreen();
	bool LoadFromFile( const uchar* File );
	bool LoadFromFileInMem( const void* File, ulong Size );
	bool LoadFromRaw( const uchar* TexData, ulong TexDataSize, ulong Width, 
		ulong Height, GLint  OpenGLComponents, GLenum OpenGLFormat, 
		GLenum OpenGLType );
	bool Reset();
    //@}

    //-------------------------------------------------------------------------
    /** \name OpenGL functions */
    //-------------------------------------------------------------------------
    //@{
	void UploadToGL();
	void Bind();
    //@}

    //-------------------------------------------------------------------------
    /** \name Get/Set functions */
    //-------------------------------------------------------------------------
    //@{
	inline bool GetUseMipMap() const { return m_UseMipMap; }
	inline TTileMode GetTileModeS() const { return m_TileModeS; }
	inline TTileMode GetTileModeT() const { return m_TileModeT; }
	inline TEnvMode GetEnvMode() const { return m_EnvMode; }
	inline TFiltering GetFiltering() const { return m_Filtering; }
	inline long GetCubeMapIndex() const { return m_CubeMapIndex; }
	inline const string& GetImagePath() const { return m_ImagePath; }
	inline ulong GetNbComponents() const { return m_OpenGLComponents; }
	inline bool	HasChanged() const { return m_HasChanged; }

	inline void SetUseMipMap( bool UseMipMap ) 
	{
		if( UseMipMap != m_UseMipMap ) 
		{
			m_HasChanged = true;
			m_UseMipMap = UseMipMap;
		}
	}
	inline void SetEnvMode( TEnvMode EnvMode ) 
	{
		if( EnvMode != m_EnvMode ) 
		{
			m_HasChanged = true;
			m_EnvMode = EnvMode;
		}
	}
	inline void SetTileMode( TTileMode TileMode ) 
	{
		if( TileMode != m_TileModeS || TileMode != m_TileModeT ) 
		{
			m_HasChanged = true;
			m_TileModeS = TileMode;
			m_TileModeT = TileMode;
		}
	}
	inline void SetTileModeS( TTileMode TileMode )
	{
		if( TileMode != m_TileModeS ) 
		{
			m_HasChanged = true;
			m_TileModeS = TileMode;
		}
	}
	inline void SetTileModeT( TTileMode TileMode )
	{
		if( TileMode != m_TileModeT ) 
		{
			m_HasChanged = true;
			m_TileModeT = TileMode;
		}
	}
	inline void SetFiltering( TFiltering Filtering )
	{
		if( Filtering != m_Filtering ) 
		{
			m_HasChanged = true;
			m_Filtering = Filtering;
		}
	}
	inline void SetCubeMapIndex( long CubeMapIndex )
	{
		if( CubeMapIndex != m_CubeMapIndex ) 
		{
			m_HasChanged = true;
			m_CubeMapIndex = CubeMapIndex;
		}
	}
	inline void SetImagePath( const string& Path )
	{
		m_ImagePath = Path;
	}
   //@}


private:

	uchar*			m_TexData;	///< The texture pixels
	ulong			m_DataSize;
	ulong			m_Width;
	ulong			m_Height;
	bool			m_HasChanged;

	string			m_ImagePath;

	TEnvMode		m_EnvMode;
	TTileMode		m_TileModeS;
	TTileMode		m_TileModeT;
	TFiltering		m_Filtering;
	bool			m_UseMipMap;
	long			m_CubeMapIndex;
	
	GLint			m_OpenGLComponents;
	GLenum			m_OpenGLFormat;
	GLenum			m_OpenGLType;
	GLuint			m_OpenGLId;


};

typedef vector<CTexture*> T_TextureVector;
typedef T_TextureVector::iterator IT_TextureVector;
typedef T_TextureVector::const_iterator CIT_TextureVector;

} // namespace Clouds

#endif // _Texture_h_
